Intel IOMMU: IOVA allocation and management routines
[pandora-kernel.git] / ipc / sem.c
index 446c8f5..35952c0 100644 (file)
--- a/ipc/sem.c
+++ b/ipc/sem.c
@@ -80,7 +80,7 @@
 #include <linux/audit.h>
 #include <linux/capability.h>
 #include <linux/seq_file.h>
-#include <linux/mutex.h>
+#include <linux/rwsem.h>
 #include <linux/nsproxy.h>
 
 #include <asm/uaccess.h>
 #define sem_ids(ns)    (*((ns)->ids[IPC_SEM_IDS]))
 
 #define sem_unlock(sma)                ipc_unlock(&(sma)->sem_perm)
-#define sem_checkid(ns, sma, semid)    \
-       ipc_checkid(&sem_ids(ns),&sma->sem_perm,semid)
-#define sem_buildid(ns, id, seq) \
-       ipc_buildid(&sem_ids(ns), id, seq)
+#define sem_checkid(sma, semid)        ipc_checkid(&sma->sem_perm, semid)
+#define sem_buildid(id, seq)   ipc_buildid(id, seq)
 
 static struct ipc_ids init_sem_ids;
 
@@ -148,7 +146,7 @@ void sem_exit_ns(struct ipc_namespace *ns)
        int next_id;
        int total, in_use;
 
-       mutex_lock(&sem_ids(ns).mutex);
+       down_write(&sem_ids(ns).rw_mutex);
 
        in_use = sem_ids(ns).in_use;
 
@@ -160,7 +158,7 @@ void sem_exit_ns(struct ipc_namespace *ns)
                freeary(ns, sma);
                total++;
        }
-       mutex_unlock(&sem_ids(ns).mutex);
+       up_write(&sem_ids(ns).rw_mutex);
 
        kfree(ns->ids[IPC_SEM_IDS]);
        ns->ids[IPC_SEM_IDS] = NULL;
@@ -174,6 +172,22 @@ void __init sem_init (void)
                                IPC_SEM_IDS, sysvipc_sem_proc_show);
 }
 
+/*
+ * This routine is called in the paths where the rw_mutex is held to protect
+ * access to the idr tree.
+ */
+static inline struct sem_array *sem_lock_check_down(struct ipc_namespace *ns,
+                                               int id)
+{
+       struct kern_ipc_perm *ipcp = ipc_lock_check_down(&sem_ids(ns), id);
+
+       return container_of(ipcp, struct sem_array, sem_perm);
+}
+
+/*
+ * sem_lock_(check_) routines are called in the paths where the rw_mutex
+ * is not held.
+ */
 static inline struct sem_array *sem_lock(struct ipc_namespace *ns, int id)
 {
        struct kern_ipc_perm *ipcp = ipc_lock(&sem_ids(ns), id);
@@ -228,6 +242,14 @@ static inline void sem_rmid(struct ipc_namespace *ns, struct sem_array *s)
  */
 #define IN_WAKEUP      1
 
+/**
+ * newary - Create a new semaphore set
+ * @ns: namespace
+ * @params: ptr to the structure that contains key, semflg and nsems
+ *
+ * Called with sem_ids.rw_mutex held (as a writer)
+ */
+
 static int newary(struct ipc_namespace *ns, struct ipc_params *params)
 {
        int id;
@@ -261,14 +283,14 @@ static int newary(struct ipc_namespace *ns, struct ipc_params *params)
        }
 
        id = ipc_addid(&sem_ids(ns), &sma->sem_perm, ns->sc_semmni);
-       if(id == -1) {
+       if (id < 0) {
                security_sem_free(sma);
                ipc_rcu_putref(sma);
-               return -ENOSPC;
+               return id;
        }
        ns->used_sems += nsems;
 
-       sma->sem_perm.id = sem_buildid(ns, id, sma->sem_perm.seq);
+       sma->sem_perm.id = sem_buildid(id, sma->sem_perm.seq);
        sma->sem_base = (struct sem *) &sma[1];
        /* sma->sem_pending = NULL; */
        sma->sem_pending_last = &sma->sem_pending;
@@ -281,6 +303,9 @@ static int newary(struct ipc_namespace *ns, struct ipc_params *params)
 }
 
 
+/*
+ * Called with sem_ids.rw_mutex and ipcp locked.
+ */
 static inline int sem_security(struct kern_ipc_perm *ipcp, int semflg)
 {
        struct sem_array *sma;
@@ -289,6 +314,9 @@ static inline int sem_security(struct kern_ipc_perm *ipcp, int semflg)
        return security_sem_associate(sma, semflg);
 }
 
+/*
+ * Called with sem_ids.rw_mutex and ipcp locked.
+ */
 static inline int sem_more_checks(struct kern_ipc_perm *ipcp,
                                struct ipc_params *params)
 {
@@ -514,9 +542,9 @@ static int count_semzcnt (struct sem_array * sma, ushort semnum)
        return semzcnt;
 }
 
/* Free a semaphore set. freeary() is called with sem_ids.mutex locked and
* the spinlock for this semaphore set hold. sem_ids.mutex remains locked
* on exit.
  */
 static void freeary(struct ipc_namespace *ns, struct sem_array *sma)
 {
@@ -601,7 +629,7 @@ static int semctl_nolock(struct ipc_namespace *ns, int semid, int semnum,
                seminfo.semmnu = SEMMNU;
                seminfo.semmap = SEMMAP;
                seminfo.semume = SEMUME;
-               mutex_lock(&sem_ids(ns).mutex);
+               down_read(&sem_ids(ns).rw_mutex);
                if (cmd == SEM_INFO) {
                        seminfo.semusz = sem_ids(ns).in_use;
                        seminfo.semaem = ns->used_sems;
@@ -610,7 +638,7 @@ static int semctl_nolock(struct ipc_namespace *ns, int semid, int semnum,
                        seminfo.semaem = SEMAEM;
                }
                max_id = ipc_get_maxid(&sem_ids(ns));
-               mutex_unlock(&sem_ids(ns).mutex);
+               up_read(&sem_ids(ns).rw_mutex);
                if (copy_to_user (arg.__buf, &seminfo, sizeof(struct seminfo))) 
                        return -EFAULT;
                return (max_id < 0) ? 0: max_id;
@@ -881,7 +909,7 @@ static int semctl_down(struct ipc_namespace *ns, int semid, int semnum,
                if(copy_semid_from_user (&setbuf, arg.buf, version))
                        return -EFAULT;
        }
-       sma = sem_lock_check(ns, semid);
+       sma = sem_lock_check_down(ns, semid);
        if (IS_ERR(sma))
                return PTR_ERR(sma);
 
@@ -962,45 +990,15 @@ asmlinkage long sys_semctl (int semid, int semnum, int cmd, union semun arg)
                return err;
        case IPC_RMID:
        case IPC_SET:
-               mutex_lock(&sem_ids(ns).mutex);
+               down_write(&sem_ids(ns).rw_mutex);
                err = semctl_down(ns,semid,semnum,cmd,version,arg);
-               mutex_unlock(&sem_ids(ns).mutex);
+               up_write(&sem_ids(ns).rw_mutex);
                return err;
        default:
                return -EINVAL;
        }
 }
 
-static inline void lock_semundo(void)
-{
-       struct sem_undo_list *undo_list;
-
-       undo_list = current->sysvsem.undo_list;
-       if (undo_list)
-               spin_lock(&undo_list->lock);
-}
-
-/* This code has an interaction with copy_semundo().
- * Consider; two tasks are sharing the undo_list. task1
- * acquires the undo_list lock in lock_semundo().  If task2 now
- * exits before task1 releases the lock (by calling
- * unlock_semundo()), then task1 will never call spin_unlock().
- * This leave the sem_undo_list in a locked state.  If task1 now creats task3
- * and once again shares the sem_undo_list, the sem_undo_list will still be
- * locked, and future SEM_UNDO operations will deadlock.  This case is
- * dealt with in copy_semundo() by having it reinitialize the spin lock when 
- * the refcnt goes from 1 to 2.
- */
-static inline void unlock_semundo(void)
-{
-       struct sem_undo_list *undo_list;
-
-       undo_list = current->sysvsem.undo_list;
-       if (undo_list)
-               spin_unlock(&undo_list->lock);
-}
-
-
 /* If the task doesn't already have a undo_list, then allocate one
  * here.  We guarantee there is only one thread using this undo list,
  * and current is THE ONE
@@ -1061,9 +1059,9 @@ static struct sem_undo *find_undo(struct ipc_namespace *ns, int semid)
        if (error)
                return ERR_PTR(error);
 
-       lock_semundo();
+       spin_lock(&ulp->lock);
        un = lookup_undo(ulp, semid);
-       unlock_semundo();
+       spin_unlock(&ulp->lock);
        if (likely(un!=NULL))
                goto out;
 
@@ -1086,10 +1084,10 @@ static struct sem_undo *find_undo(struct ipc_namespace *ns, int semid)
        new->semadj = (short *) &new[1];
        new->semid = semid;
 
-       lock_semundo();
+       spin_lock(&ulp->lock);
        un = lookup_undo(ulp, semid);
        if (un) {
-               unlock_semundo();
+               spin_unlock(&ulp->lock);
                kfree(new);
                ipc_lock_by_ptr(&sma->sem_perm);
                ipc_rcu_putref(sma);
@@ -1100,7 +1098,7 @@ static struct sem_undo *find_undo(struct ipc_namespace *ns, int semid)
        ipc_rcu_putref(sma);
        if (sma->sem_perm.deleted) {
                sem_unlock(sma);
-               unlock_semundo();
+               spin_unlock(&ulp->lock);
                kfree(new);
                un = ERR_PTR(-EIDRM);
                goto out;
@@ -1111,7 +1109,7 @@ static struct sem_undo *find_undo(struct ipc_namespace *ns, int semid)
        sma->undo = new;
        sem_unlock(sma);
        un = new;
-       unlock_semundo();
+       spin_unlock(&ulp->lock);
 out:
        return un;
 }
@@ -1287,10 +1285,6 @@ asmlinkage long sys_semop (int semid, struct sembuf __user *tsops, unsigned nsop
 
 /* If CLONE_SYSVSEM is set, establish sharing of SEM_UNDO state between
  * parent and child tasks.
- *
- * See the notes above unlock_semundo() regarding the spin_lock_init()
- * in this code.  Initialize the undo_list->lock here instead of get_undo_list()
- * because of the reasoning in the comment above unlock_semundo.
  */
 
 int copy_semundo(unsigned long clone_flags, struct task_struct *tsk)
@@ -1356,7 +1350,7 @@ void exit_sem(struct task_struct *tsk)
                if (u->semid == -1)
                        goto next_entry;
 
-               BUG_ON(sem_checkid(ns,sma,u->semid));
+               BUG_ON(sem_checkid(sma, u->semid));
 
                /* remove u from the sma->undo list */
                for (unp = &sma->undo; (un = *unp); unp = &un->id_next) {
@@ -514,9 +542,9 @@ static int count_semzcnt (struct sem_array * sma, ushort semnum)
 }
 
-/* Free a semaphore set. freeary() is called with sem_ids.mutex locked and
- * the spinlock for this semaphore set hold. sem_ids.mutex remains locked
- * on exit.
+/* Free a semaphore set. freeary() is called with sem_ids.rw_mutex locked
+ * as a writer and the spinlock for this semaphore set hold. sem_ids.rw_mutex
+ * remains locked on exit.
  */
 static void freeary(struct ipc_namespace *ns, struct sem_array *sma)